
#include "qobjects.h"

#include <cstdlib>
#include <ctime>

int main(int argc, char* argv[]){

	//---------------------------------------initialize parameters (former seed 3156)

	const double hbar = 1,                // fundamental constants 
				 e  = 2.7182818,
				 pi = 3.1415927;
		
	const std::complex<double> ei(0,1);

	const unsigned int N = 10,             // number of fock states
					   ns = 100000;        // k=0.1:  50k, 100k 
	
	bool double_noise = false;      // "true" for testing strong convergence 
    
	const int          nout = 20;       

	const double T = 10,              // k=0.1: 10             // total time 
				 dt = T/ns,
				 
				 w = 2*pi, 
				 g = 0.1;             // noise strength 
				 
	const int	 ksteps = 1; 
	std::vector<double> k_vec(ksteps); k_vec[0] = 0.1;   // 4 gives nice jumps 

	const int	 seed[ksteps] = {931932};    //  k=0.1: 931932    k=4: 4132  
		
	//--------------------------------initialise files

	std::ofstream series("series_k01.dat",std::ios_base::trunc); 
	  
	//--------------------------------initialize operators 
					
	qoperator I("identity",N), 
			  a("destroy",N);

	qoperator adag = Hconj(a), 
			  E = adag*a,  
			  x = std::sqrt(0.5)*(a + adag),  
			  x2 = x*x; 
		
	// free Hamiltonian 
	qoperator H = w*E;
	qoperator dU = (-ei*dt)*H; 
		
	// observable to measure
	qoperator M  = E;
	qoperator M2 = E*E;
	  
	// observable to display   
	qoperator O1 = E;

	//initial states 
	//qmatrix rho0("thermal",3.0,N); 
	qstate psi0("number",3.0,N); 
	qmatrix rho0(psi0);
			
	//------------------------------- set up k vector 

	//if (ksteps>1) {
	//	double one = 1; // avoid compiler warning 
	//	double dk = (std::log(kmax) - std::log(kmin))/(ksteps-one);
	//	for (int m=0;m<ksteps;m++) { k_vec[m] = std::exp(std::log(kmin) + m*dk); } 
	//}
	//else { k_vec[0] = kmin; } 

	//------------------------------------------------------------
	//                        Simulation
	//------------------------------------------------------------
		
	//---------------------------------timing initialisation
		
	time_t time0, *time_ptr = 0;       
	time0 = time(time_ptr);
		

	//------------------------------ loop over values of k and trajectories

	for (int m=0;m<ksteps;m++) { 

		double k = k_vec[m];   
		double r8kdt = std::sqrt(8*k*dt); 
		
		//noise initialization 
		std::vector<double> m_noise(ns);
		std::srand(seed[m]); 
		if (double_noise) {
			std::vector<double> d_noise(2*ns);
			make_noise(d_noise);
			for (unsigned int i=0;i<ns;i++) { m_noise[i] = std::sqrt(0.5)*(d_noise[2*i] + d_noise[2*i+1]); }
			for (unsigned int i=0;i<ns;i++) { m_noise[i] /= r8kdt; }
		} else {
			make_noise(m_noise); 
			for (unsigned int i=0;i<ns;i++) { m_noise[i] /= r8kdt; }
		}
		// save noise in file for wfmc to load 
//		std::ofstream noisefl("noisefl_100k.dat");
//		for (unsigned int i=0;i<ns;i++) { noisefl << m_noise[i] << std::endl; }
//		noisefl.close(); 
		
		std::cout << "noise: " << m_noise[0] << " " << m_noise[1] << std::endl;

		//initial state 
		qmatrix rho = rho0;   
	  
		double EM = trace(M*rho).real(),  // initialise variables
			   EO1 = trace(O1*rho).real(),
			   d_rec, dW;          
		
		// integrate a single trajectory
		for (unsigned int i=0;i<ns;i++) {  
							
			// output <E> time series  
			if (std::div(i,nout).rem==0) { series << EO1 << " "; } 
		  
			// report progress
			if (std::div(i,10000).rem==0) {
				std::cout << i << ": " << EO1 << " " << "T = " << i*dt << std::endl;
			} 
							
			// Measurement Evolution 
			EM = trace(M*rho).real();
			d_rec = 4*k*dt*(EM + m_noise[i]); 
			dW = m_noise[i]*std::sqrt(8*k)*dt; 
			rho += (-k*dt)*(M2*rho + rho*M2 - 2.0*M*rho*M) + d_rec*(M*rho + rho*M)
				   + (k*(dW*dW - dt))*(M2*rho + rho*M2 + 2*(M*rho*M));
		
			//Free Hamiltonian evolution (1st-order in dt)
			rho  += dU*rho - rho*dU;
			
			// Noise Evolution (1st order in dt)
			rho += (-(g/2)*dt)*(x2*rho + rho*x2 - 2.0*x*rho*x);
			rho.normalise(); 		

			// expectation value to display  
			EO1 = trace(O1*rho).real();
														
		} //i 
		
		series << EO1 << std::endl;
			
	} //m  
		
	//----------------------- output simulation time

	double took = difftime(time(time_ptr),time0);
	std::cout << std::endl << "Simulation took " << took << " seconds" << std::endl << std::endl;
		

	//----------------------- tidy up

	series.close();
					 
	return 0;
}
