
#include "qobjects.h"
#include "mpi.h" 
#include <ctime>

//-----------------------------------------------------------------------------------------------
int thermal_ens(std::vector<qstate>& psi, double avn, int N); 
double global_mean( const std::vector<qstate> &psi_vec, std::vector<double> &P, qoperator &Op); 
int regenerate(std::vector<qstate>& psi, 
			   std::vector<double>& P, 
			   double drop_thresh, double& min_Ens, double& av_Ens, double& max_Pdrop, 
			   int my_rank, int nproc);   
//-----------------------------------------------------------------------------------------------

int main(int argc, char* argv[]){

	//---------------------------------------initialize MPI 
		
	int my_rank, nproc; 
	MPI_Status status;
	MPI_Init(&argc,&argv);
	MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
	MPI_Comm_size(MPI_COMM_WORLD,&nproc);
		
	//---------------------------------------initialize parameters 
		
	const int	 seed = 931932; 
	const double hbar = 1,                 // fundamental constants 
				 e  = 2.7182818,
				 pi = 3.1415927;
	const std::complex<double> ei(0,1);

	const unsigned int N = 10,             // number of fock states
					   ns = 100000,        // k01: 50,000   k01t2: 100,000 
					   nreg = 10;  
	const int          nout = 20;
	
	bool double_noise = false;      // for testing strong convergence 

	const double T = 10,              // total time 
				 dt = T/ns,
				 
				 w = 2*pi, k = 0.1, g = 0.1, avn = 3;  
							
	//--------------------------------initialise files
	std::ofstream series("series_k01.dat"); 
					 
	//------------------------------- make measurement noise  (or load from file below)
	std::vector<double> m_noise(ns); 
	std::srand(seed); 
	const double r8kdt = std::sqrt(8*k*dt); 
	make_noise(m_noise); 
	for (unsigned int i=0;i<ns;i++) { m_noise[i] /= r8kdt; }
	
	//------------------------------- load meas noise from file 
//	std::vector<double> m_noise(ns); 
//	if (my_rank==0) { 
//		if (double_noise) { 
//			std::vector<double> d_noise(2*ns);
//			std::ifstream noisefl("noisefl_100k.dat"); 
//			for (unsigned int i=0;i<2*ns;i++) { noisefl >> d_noise[i]; } 
//			noisefl.close(); 
//			for (unsigned int i=0;i<ns;i++) { m_noise[i] = 0.5*(d_noise[2*i] + d_noise[2*i+1]); }
//		}
//		else {
//			std::ifstream noisefl("noisefl_100k.dat"); 
//			for (unsigned int i=0;i<ns;i++) { noisefl >> m_noise[i]; } 
//			noisefl.close(); 
//		}
//	}
	
	MPI_Bcast(&(m_noise[0]),ns,MPI_DOUBLE,0,MPI_COMM_WORLD); 
	
	if (my_rank==0) { std::cout << "noise: " << m_noise[0] << " " << m_noise[1] << std::endl; }
	
	//-------------------------------- Monte-Carlo set-up   
	const int tpp = 16, tot_traj = tpp*nproc;  // (tpp: 16 with 64 procs gives 1024)
	const double meanP = 1.0/tot_traj, drop_thresh = 0.2*meanP;  // < 0.7*meanP 
	double min_Ens(tpp*nproc), av_Ens(0), max_Pdrop(0); 
	std::vector<double> mc_noise(2*tpp); 
	std::srand(seed + 1000*my_rank); 		
				
	//--------------------------------initialize operators 
	qoperator I("identity",N), a("destroy",N), ad = Hconj(a), 
	E = ad*a, x = std::sqrt(0.5)*(a + ad), x2 = x*x; 
	qoperator H = w*E, dU = (-ei*dt)*H; // free Hamiltonian 
	qoperator M  = E, M2 = E*E;			// observable to measure 
	qoperator O1 = E;					// observable to display
		
	//------------------------------- initial state 
	std::vector<qstate> psi(tpp);      // states 
	std::vector<double> P(tpp,meanP);  // ensemble probabilities  
	//thermal_ens(psi, avn, N); 
	qstate psi0("number",3,N); 
	for (int m=0; m<tpp; m++) { psi[m] = psi0; }
		
	//---------------------------------timing initialisation
	time_t time0, *time_ptr = 0;       
	time0 = time(time_ptr);

	double EM = global_mean(psi,P,M),  // initialise values 
		   EO1 = global_mean(psi,P,O1), 
		   dW;        
	const double rgdt = std::sqrt(g*dt);
			
	//------------------------------------------------------------
	//                        Simulation
	//------------------------------------------------------------
		
	for (unsigned int i=0;i<ns;i++) {  
					
		if (my_rank==0) { 
			if (std::div(i,nout).rem==0) { series << EO1 << " "; } // output <E> time series
			if (std::div(i,100).rem==0) { std::cout << i << ": " << EO1 << " " << "T = " << i*dt << std::endl; } 
		}
				
		EM = global_mean(psi,P,M); 
		double d_rec = 4*k*dt*(EM + m_noise[i]);
		
		make_noise(mc_noise); 
		
		for (int m=0; m<tpp; m++) {			
			// Measurement Evolution 
			dW = m_noise[i]*std::sqrt(8*k)*dt; 
			psi[m]  += (-k*dt)*(M2*psi[m]) + d_rec*(M*psi[m]) + (k*(dW*dW - dt))*(M2*psi[m]);  
			P[m] = P[m]*psi[m].norm(); 
			
			//Free Hamiltonian evolution (1st-order in dt) 
			psi[m]  += dU*psi[m];
 
			// Noise Evolution 
			psi[m]  += (-(g/2)*dt)*(x2*psi[m]) + (ei*rgdt*mc_noise[m])*(x*psi[m]) 
			           - ((dt*g/2)*(1 - mc_noise[m]*mc_noise[m]))*(x2*psi[m]);  
			
			psi[m].normalise();
		}
				
		// normalize the weightings 
		double sub_Pnorm(0), Pnorm(0);
		for (int m=0;m<tpp;m++) { sub_Pnorm += P[m]; }
		MPI_Allreduce(&sub_Pnorm,&Pnorm,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);	
		for (int m=0;m<tpp;m++) { P[m] /= Pnorm; } 
		
		// regenerate the ensemble 
		if (std::div(i,nout).rem==0) { 
			regenerate(psi,P,drop_thresh,min_Ens,av_Ens,max_Pdrop,my_rank,nproc); 
		}
				
		EO1 = global_mean(psi,P,O1);  // expectation value to display
															
	} //i 

	if (my_rank==0) { 
		series << EO1 << std::endl;
		//----------------------- report ensemble stats 
		std::cout << std::endl << "Min Ensemble size: " << min_Ens << std::endl; 
		std::cout << std::endl << "Ave Ensemble size: " << av_Ens << std::endl;
		std::cout << std::endl << "Max Pdrop: " << max_Pdrop << std::endl;
		//----------------------- report simulation time
		std::cout << std::endl << "Simulation took " << difftime(time(time_ptr),time0)/60 << "mins" 
				  << std::endl << std::endl;
	}
		
	//----------------------- tidy up 
	series.close();  
	MPI_Finalize(); 
					 
	return 0; 
}


// -----------------------------------------------------------------------
// ---------------------------- procedures -------------------------------

int thermal_ens(std::vector<qstate>& psi, double avn, int N) {
	unsigned int tpp = psi.size(); 
	double s = avn/(1 + avn);
	std::vector<double> thermal_dens(N); thermal_dens[0] = 1; 
	double norm = thermal_dens[0];
	for (unsigned int i=1;i<N;++i) { 
		thermal_dens[i] = s*thermal_dens[i-1]; 
		norm += thermal_dens[i];  
	} 
	for (unsigned int i=0;i<N;i++) { thermal_dens[i] /= norm; }
	std::vector<double> thermal_dist(N); thermal_dist[0] = thermal_dens[0]; 
    for (unsigned int i=1;i<N;i++) { thermal_dist[i] = thermal_dist[i-1] + thermal_dens[i]; }
	for (unsigned int m=0;m<tpp;m++) {
		int n_pick = sample(thermal_dist);
		qstate psi_number("number",n_pick,N); 
		psi[m] = psi_number; 
	}
	return 0; 
}

double global_mean( const std::vector<qstate> &psi_vec, std::vector<double> &P, qoperator &Op)
{
	double sub_meanM(0), meanM(0); 
	int traj_per_proc = psi_vec.size(); 
	for (int traj=0;traj<traj_per_proc;traj++) { 
		sub_meanM += P[traj]*(Op.expect(psi_vec[traj])).real(); 
	}
	MPI_Allreduce(&sub_meanM,&meanM,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
	return meanM; 
}

int regenerate(std::vector<qstate>& psi, 
			   std::vector<double>& P, 
			   double drop_thresh, double& min_Ens, double& av_Ens, double& max_Pdrop, 
			   int my_rank, int nproc)
{
	MPI_Status status; 
	
	static long unsigned int count(0); 
	count++; 
	int tpp = psi.size(); 
	int tot_traj = nproc*tpp; 
	unsigned int state_size = psi[0].len(); 
	
	// gather all the Ps 
	std::vector<double> P_all(tot_traj), D_all(tot_traj); 
	MPI_Allgather(&(P[0]),tpp,MPI_DOUBLE,&(P_all[0]),tpp,MPI_DOUBLE,MPI_COMM_WORLD);
	
	// form cumulative distribution 
	D_all[0] = P_all[0];
	for (int q=1;q<tot_traj;q++) { D_all[q] = D_all[q-1] + P_all[q]; }
	
	// pick which states to replace
	std::vector<int> Targ_all(tot_traj,0), Source_all(tot_traj,0);
	if (my_rank==0) { 
		for (int q=0;q<tot_traj;q++) { 
			if (P_all[q] < drop_thresh) { 
				Targ_all[q] = 1;  
				Source_all[q] = sample(D_all);
			}
		} 
	}
	MPI_Bcast(&(Targ_all[0]),tot_traj,MPI_INT,0,MPI_COMM_WORLD);
	MPI_Bcast(&(Source_all[0]),tot_traj,MPI_INT,0,MPI_COMM_WORLD); 
	
	// calculate Pdrop and Ens size 
	double Pdrop(0);
	for (int q=0;q<tot_traj;q++) { if (Targ_all[q]==1) { Pdrop += P_all[q]; } } 
	if (max_Pdrop < Pdrop) { max_Pdrop = Pdrop; }
	double sub_Entsum(0), Entsum(0);
	for (int traj=0;traj<tpp;traj++) { sub_Entsum -= P[traj]*std::log(P[traj]); }
	MPI_Reduce(&sub_Entsum,&Entsum,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
	double Ens_size = std::exp(Entsum);
	av_Ens *= static_cast<double>(count-1);  
	av_Ens += Ens_size; 
	av_Ens /= static_cast<double>(count);
	if (min_Ens > Ens_size) { min_Ens = Ens_size; }
	
	// Replace the states (NB: all processors do this, so that they all have the info)
	for (int q=0;q<tot_traj;q++) {  
		if (Targ_all[q] == 1) { 
			int target_proc = std::div(q,tpp).quot;
			int target_j = std::div(q,tpp).rem;
			if (Source_all[q] != q) { 
				int source_proc = std::div(Source_all[q],tpp).quot;
				int source_j = std::div(Source_all[q],tpp).rem; 
				if (source_proc == target_proc) {
					if (my_rank == source_proc) { 
						P[source_j] /= 2;
						psi[target_j] = psi[source_j]; 
						P[target_j] = P[source_j];
					}
				} else {
					if (my_rank == source_proc) {
						P[source_j] /= 2; 
						MPI_Send(psi[source_j].get_array(),2*state_size,MPI_DOUBLE,target_proc,2*q,MPI_COMM_WORLD);
						MPI_Send(&(P[source_j]),1,MPI_DOUBLE,target_proc,2*q+1,MPI_COMM_WORLD); 
					}
					if (my_rank == target_proc) {
						MPI_Recv(psi[target_j].get_array(),2*state_size,MPI_DOUBLE,source_proc,2*q,MPI_COMM_WORLD,&status);
						MPI_Recv(&(P[target_j]),1,MPI_DOUBLE,source_proc,2*q+1,MPI_COMM_WORLD,&status); 
					}				 
				} 
			} 
		}
	}
	
	// normalize the weightings 
	double Pnorm(0), Ptot(0);
	for (int j=0;j<tpp;j++) { Pnorm += P[j]; }
	MPI_Allreduce(&Pnorm,&Ptot,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);	
	for (int j=0;j<tpp;j++) { P[j] /= Ptot; }		
	
	return 0; 
}


